/*
 * This program tests to guarantee that different vcpus in a virtual machine
 * don't end up sharing the TLB entries.
 *
 * VT-x implementations can cache guest linear -> host physical mappings in
 * TLB. These mappings are tagged with the EP4TA and the VPID. Thus it is
 * possible for multiple vcpus that share the EPT to also share a host cpu
 * since each vcpu's mappings can be identified with the associated VPID.
 *
 * This program works best if multiple vcpus belonging to a virtual machine
 * share the same host cpu:
 *
 * For e.g., create a 2 vcpu virtual machine 'vm1' executing on host cpu 0:
 * sudo cpuset -l 0 ./vmrun.sh -c 2 vm1
 *
 * Now create two or more 'vpid' processes within the guest:
 * ./vpid &
 * ./vpid &
 * ./vpid &
 * ./vpid &
 *
 * If there is any inadvertent sharing of combined mappings between the vcpus
 * this program will assert() because the value read from the mmap'ed region
 * will not match the value returned by 'getpid()'.
 */

#include <sys/param.h>
#include <sys/mman.h>
#include <unistd.h>
#include <assert.h>

#define MMAP_ADDR       ((void *)(uintptr_t)(126UL << 39))
#define MMAP_LEN	4096

int
main(int argc, char *argv[])
{
	int i;
	int *ip;

	ip = mmap(MMAP_ADDR, MMAP_LEN, PROT_READ|PROT_WRITE,
		  MAP_ANON | MAP_FIXED, -1, 0);
	assert(ip != MAP_FAILED);

	for (i = 0; i < MMAP_LEN / sizeof(int); i++)
		ip[i] = getpid();

	while (1) {
		for (i = 0; i < MMAP_LEN / sizeof(int); i++) {
			assert(ip[i] == getpid());
		}
	}
}
